import greenfoot.*;
import java.util.List;
import java.util.ArrayList;

public abstract class ThreeD extends Actor
{
    double x, y, z, h, w, d;
    String imgName = "";
    static List<Actor> actors = new ArrayList();

    public ThreeD(double horizontal, double vertical, double distance, double high, double wide, String iName)
    {
        x = horizontal; // x offset from viewpoint in viewplane
        y = vertical; // y offset from viewpoint in viewplane
        z = distance; // z offset from xy-offset perpendicular to viewplane
        h = high; // height of image (not actual)
        w = wide; // width of image (not actual)
        d = Math.sqrt(Math.pow(x, 2)+Math.pow(y, 2)+Math.pow(z, 2));
        imgName = iName; // String name of image
        move();
        actors.add(this); // populates actors ArrayList object
    }

    public void move()
    {
        double plnD = Math.hypot(x, y); // distance of xy-offset from viewpoint
        double dist = Math.hypot(plnD, z); // distance of object from viewpoint
        double hwFactor = 400/dist; // factor used to determine visual height and width
        double xyFactor = 400/z; // factor used to determine location on viewplane
        int viewX = (int)(x*xyFactor);
        int viewY = (int)(y*xyFactor);
        setLocation((int)viewX+Scene.width/2, (int)viewY+Scene.height/4);
        if (Math.abs(viewX) <= Scene.width/2 && Math.abs(viewY-Scene.height/4) <= Scene.height/2)
        {
            GreenfootImage img = new GreenfootImage(imgName);
            img.scale((int)(h*hwFactor), (int)(w*hwFactor));
            setImage(img);
        }
    }

    public static void fixLocations(World w)
    {
        int c = actors.size();
        // the following sorts the actors from farthest to nearest
        if (c < 2) return;
        int[] order = new int[c];
        for (int i=0; i<c; i++) order[i] = i;
        for (int i=0; i<c-1; i++)
        {
            int i2 = c-i-2;
            while (i2 < c-1 && ((ThreeD)actors.get(order[i2])).z < ((ThreeD)actors.get(order[i2 + 1])).z)
            {
                int hold = order[i2];
                order[i2] = order[i2+1];
                order[i2+1] = hold;
                i2++;
            }
        }
        // the following displays the properly overlapping order of actors
        for (int i=0; i<c; i++)
        {
            ThreeD a = (ThreeD)actors.get(order[i]);
            w.removeObject(a);
            w.addObject(a, 0, 0);
            a.move();
        }
    }

    public static void moveEye(World w, double dx, double dy, double dz)
    {
        int c = actors.size();
        // the following checks to make sure no objects are closer than a distance of 100
        for (int i=0; i<c; i++)
        {
            ThreeD a = (ThreeD)actors.get(i);
            if (Math.abs(a.z+dz) < 100) return;
        }
        // the following adjusts the location values of all the actors in the world
        for (int i=0; i<c; i++)
        {
            ThreeD a = (ThreeD)actors.get(i);
            a.x += dx;
            a.y += dy;
            a.z += dz;
            a.d = Math.sqrt(Math.pow(a.x, 2)+Math.pow(a.y, 2)+Math.pow(a.z, 2));
        }
    }

    // The turnEye method is a makeshift cheat in the actual action
    // Multiple usage will cause the objects to move some with respect to each other
    // True code would require changing the viewplane which in turn would
    // cause all x and z values to change while keeping the distance between
    // the object and the viewpoint the same (not an easy task)
    // maybe I will figure it out sometime (lots of Maths!!)
    public static void turnEye(World w, int dir)
    {
        int c = actors.size();
        for (int i=0; i<c; i++)
        {
            ThreeD a = (ThreeD)actors.get(i);
            double az = a.z, ax = a.x;
            a.x += az*Math.sin((double)dir*.05);
            a.z -= ax*Math.sin((double)dir*.05);
        }
    }
}